from pathlib import Path
import subprocess


def compile_with_f2py(
    files_to_compile: list[str],
    module_name: str,
    output_folder: Path,
) -> Path:
    """Compiles a list of fortran files into a module."""
    from ..shared import library_extensions

    def compile_fortran_files(quiet: bool):
        command = (
            ["python3", "-m", "numpy.f2py", "-c"]
            + files_to_compile
            + ["-m", module_name]
        )

        return subprocess.run(command, capture_output=quiet, check=not quiet)

    result = compile_fortran_files(True)

    if not result.returncode == 0:
        # If the first compile attempt fails, do it again
        # and this time print output
        compile_fortran_files(False)

    library_found = []
    for file in Path(".").iterdir():
        if file.name.startswith(module_name) and file.suffix in library_extensions:
            output_name = output_folder / f"{module_name}{file.suffix}"
            library_found.append(output_name)
            file.rename(output_name)

    assert len(library_found) > 0, f"Could not find compiled library for {module_name}"
    assert len(library_found) == 1, f"Found multiple compiled objects for {module_name}"
    return library_found[0]



"""Set up the pint library for unit handling."""

import warnings
from functools import wraps
from typing import Any, Union

import numpy as np
import numpy.typing as npt
import pint
import pint_xarray
import xarray as xr
from pint import UnitStrippedWarning, DimensionalityError

ureg = pint_xarray.setup_registry(
    pint.UnitRegistry(
        force_ndarray_like=True,
    )
)
pint.set_application_registry(ureg)

Quantity = ureg.Quantity


def suppress_downcast_warning(func):
    """Suppresses a common warning about downcasting quantities to arrays."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="The unit of the quantity is stripped when downcasting to ndarray.",
            )
            return func(*args, **kwargs)

    return wrapper


def convert_units(
    array: Union[xr.DataArray, pint.Quantity], units: Any
) -> Union[xr.DataArray, pint.Quantity]:
    """Convert an array to specified units, handling both Quantities and xr.DataArrays."""
    if isinstance(array, xr.DataArray):
        if not hasattr(array.pint, "units") or array.pint.units is None:
            array = array.pint.quantify(ureg.dimensionless)

        return array.pint.to(units)  # type: ignore[no-any-return]
    elif isinstance(array, Quantity):
        return array.to(units)  # type:ignore[no-any-return]
    else:
        raise NotImplementedError(
            f"No implementation for 'convert_units' with an array of type {type(array)} ({array})"
        )


@suppress_downcast_warning
def magnitude(
    array: Union[xr.DataArray, pint.Quantity]
) -> Union[npt.NDArray[np.float32], float]:
    """Return the magnitude of an array, handling both Quantities and xr.DataArrays."""
    if isinstance(array, xr.DataArray):
        return array.pint.dequantify()
    elif isinstance(array, Quantity):
        return array.magnitude  # type: ignore[no-any-return]
    else:
        raise NotImplementedError(
            f"No implementation for 'magnitude' with an array of type {type(array)} ({array})"
        )


def dimensionless_magnitude(
    array: Union[xr.DataArray, pint.Quantity]
) -> Union[npt.NDArray[np.float32], float]:
    """Converts the array to dimensionless and returns the magnitude."""
    return magnitude(convert_units(array, ureg.dimensionless))


__all__ = [
    "DimensionalityError",
    "UnitStrippedWarning",
    "ureg",
    "Quantity",
    "convert_units",
    "magnitude",
    "suppress_downcast_warning",
    "dimensionless_magnitude",
]



import urllib.request
import shutil
from pathlib import Path
from .compile_with_f2py import compile_with_f2py


def prepare_adas_fortran_interface(reader_dir: Path, config: dict, verbose: int):
    from ..shared import (
        fortran_file_handling_source,
        library_extensions,
    )

    fortran_file_handling_library = None
    for file in reader_dir.iterdir():
        if (
            file.name.startswith("fortran_file_handling")
            and file.suffix in library_extensions
        ):
            fortran_file_handling_library = file

    if fortran_file_handling_library is None:
        if verbose:
            print(f"Compiling {reader_dir / 'fortran_file_handling.f90'}")
        (reader_dir / "fortran_file_handling.f90").write_text(
            fortran_file_handling_source.read_text()
        )

        fortran_file_handling_library = compile_with_f2py(
            files_to_compile=[reader_dir / "fortran_file_handling.f90"],
            module_name="fortran_file_handling",
            output_folder=reader_dir,
        )
    elif verbose:
        print(f"Reusing {fortran_file_handling_library}")

    adf_file_reader = dict()
    for reader_name in config.keys():
        adf_file_reader[reader_name] = None

        (reader_dir / reader_name).mkdir(exist_ok=True, parents=True)
        for file in (reader_dir / reader_name).iterdir():
            if file.name.startswith(reader_name) and file.suffix in library_extensions:
                adf_file_reader[reader_name] = file

        if adf_file_reader[reader_name] is None:
            build_adas_file_reader(reader_dir, reader_name, verbose=verbose)
        elif verbose:
            print(f"Reusing {reader_name}_reader")


def build_adas_file_reader(
    reader_dir: Path,
    reader_name: str,
    verbose: int,
    url_base: str = "https://open.adas.ac.uk",
):
    """Builds an ADAS file reader by wrapping fortran code using f2py."""
    from ..shared import reader_pyf_source

    assert reader_name.startswith(
        "adf"
    ), f"Reader should be of the format adfXX where XX is an integer"

    reader_int = int(reader_name.lstrip("adf"))
    output_folder = reader_dir / reader_name
    (output_folder / f"xxdata_{reader_int}.pyf").write_text(
        reader_pyf_source[reader_name].read_text()
    )

    archive_file = f"xxdata_{reader_int}.tar.gz"
    query_path = f"{url_base}/code/{archive_file}"
    output_filename = output_folder / archive_file

    if not output_filename.exists():
        if verbose:
            print(f"Downloading {query_path} to {output_filename}")
        urllib.request.urlretrieve(query_path, output_filename)
        if verbose:
            print(f"Unpacking {output_filename} into {output_folder}")
        shutil.unpack_archive(output_filename, output_folder)
    else:
        if verbose:
            print(f"Reusing {query_path} ({output_filename} already exists)")

    fortran_files = [
        str(file)
        for file in (output_folder / f"xxdata_{reader_int}").iterdir()
        if (file.suffix == ".for") and not (file.stem == "test")
    ]

    fortran_files = fortran_files + [str(output_folder / f"xxdata_{reader_int}.pyf")]

    if verbose:
        print(f"Compiling {reader_name}")
    compile_with_f2py(
        files_to_compile=fortran_files,
        module_name=f"{reader_name}_reader",
        output_folder=output_folder,
    )



from pathlib import Path
import importlib.util


def load_library(library_name: str, filepath: Path):
    spec = importlib.util.spec_from_file_location(library_name, filepath)

    library = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(library)

    return library


def read_adf11_file(
    reader_dir, data_file_dir, species_name, dataset_type, dataset_config
) -> dict:
    """Open and read an ADF11 OpenADAS file.

    Key to outputs (from xxdata_11.pdf)
    type   | name       | description
    (i*4)  | iz0        | nuclear charge
    (i*4)  | is1min     | minimum ion charge + 1
           |            | (generalised to connection vector index)
    (i*4)  | is1max     | maximum ion charge + 1
           |            | (note excludes the bare nucleus)
           |            | (generalised to connection vector index and excludes
           |            | last one which always remains the bare nucleus)
    (i*4)  | nptnl      | number of partition levels in block
    (i*4)  | nptn()     | number of partitions in partition level
           |            | 1st dim: partition level
    (i*4)  | nptnc(,)   | number of components in partition
           |            | 1st dim: partition level
           |            | 2nd dim: member partition in partition level
    (i*4)  | iptnla()   | partition level label (0=resolved root,1=
           |            | unresolved root)
           |            | 1st dim: partition level index
    (i*4)  | iptna(,)   | partition member label (labelling starts at 0)
           |            | 1st dim: partition level index
           |            | 2nd dim: member partition index in partition level
    (i*4)  | iptnca(,,) | component label (labelling starts at 0)
           |            | 1st dim: partition level index
           |            | 2nd dim: member partition index in partition level
           |            | 3rd dim: component index of member partition
    (i*4)  | ncnct      | number of elements in connection vector
    (i*4)  | icnctv()   | connection vector of number of partitions
           |            | of each superstage in resolved case
           |            | including the bare nucleus
           |            | 1st dim: connection vector index
    (i*4)  | iblmx      | number of (sstage, parent, base)
           |            | blocks in isonuclear master file
    (i*4)  | ismax      | number of charge states
           |            | in isonuclear master file
           |            | (generalises to number of elements in
           |            |  connection vector)
    (c*12) | dnr_ele    | CX donor element name for iclass = 3 or 5
           |            | (blank if unset)
    (r*8)  | dnr_ams    | CX donor element mass for iclass = 3 or 5
           |            | (0.0d0 if unset)
    (i*4)  | isppr()    | 1st (parent) index for each partition block
           |            | 1st dim: index of (sstage, parent, base)
           |            |          block in isonuclear master file
    (i*4)  | ispbr()    | 2nd (base) index for each partition block
           |            | 1st dim: index of (sstage, parent, base)
           |            |          block in isonuclear master file
    (i*4)  | isstgr()   | s1 for each resolved data block
           |            | (generalises to connection vector index)
           |            | 1st dim: index of (sstage, parent, base)
           |            |          block in isonuclear master file
    (i*4)  | idmax      | number of dens values in
           |            | isonuclear master files
    (i*4)  | itmax      | number of temp values in
           |            | isonuclear master files
    (r*8)  | ddens()    | log10(electron density(cm-3)) from adf11
    (r*8)  | dtev()     | log10(electron temperature (eV) from adf11
    (r*8)  | drcof(,,)  | if(iclass <=9):
           |            | 	log10(coll.-rad. coefft.) from
           |            | 	isonuclear master file
           |            | if(iclass >=10):
           |            | 	coll.-rad. coefft. from
           |            | 	isonuclear master file
           |            | 1st dim: index of (sstage, parent, base)
           |            | 		 block in isonuclear master file
           |            | 2nd dim: electron temperature index
           |            | 3rd dim: electron density index
    (l*4)  | lres       | = .true. => partial file
           |            | = .false. => not partial file
    (l*4)  | lstan      | = .true. => standard file
           |            | = .false. => not standard file
    (l*4)  | lptn       | = .true. => partition block present
           |            | = .false. => partition block not present
    """
    fortran_file_handling = load_library(
        "fortran_file_handling", reader_dir / "fortran_file_handling.so"
    )
    adf11_reader = load_library(
        "adf11_reader", reader_dir / "adf11" / "adf11_reader.so"
    )

    filename = data_file_dir / f"{species_name}_{dataset_type}.dat"

    if not filename.exists():
        raise FileNotFoundError(f"{filename} does not exist.")

    file_unit = 10

    data = dict()

    fortran_file_handling.open_file(str(filename), file_unit)
    (
        data["iz0"],
        data["is1min"],
        data["is1max"],
        data["nptnl"],
        data["nptn"],
        data["nptnc"],
        data["iptnla"],
        data["iptna"],
        data["iptnca"],
        data["ncnct"],
        data["icnctv"],
        data["iblmx"],
        data["ismax"],
        data["dnr_ele"],
        data["dnr_ams"],
        data["isppr"],
        data["ispbr"],
        data["isstgr"],
        data["idmax"],
        data["itmax"],
        data["ddens"],
        data["dtev"],
        data["drcof"],
        data["lres"],
        data["lstan"],
        data["lptn"],
    ) = adf11_reader.xxdata_11(
        # unit to which input file is allocated
        iunit=file_unit,
        # class of data (numerical code)
        iclass=dataset_config["code"],
        #
        # Hard-coded values for ADF11 files
        #
        # maximum number of (sstage, parent, base) blocks in isonuclear master files
        isdimd=200,
        # maximum number of dens values in isonuclear master files
        iddimd=40,
        # maximum number of temp values in isonuclear master files
        itdimd=50,
        # maximum level of partitions
        ndptnl=4,
        # maximum no. of partitions in one level
        ndptn=128,
        # maximum no. of components in a partition
        ndptnc=256,
        # maximum number of elements in connection vector
        ndcnct=100,
    )

    fortran_file_handling.close_file(file_unit)

    return data
